# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from orchestrator.config.assignee import Assignee
from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.subscription import SubscriptionBaseSchema, SubscriptionSchema
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus


class ProcessIdSchema(OrchestratorBaseModel):
    id: UUID


class ProcessForm(OrchestratorBaseModel):
    title: str
    type: str
    properties: Dict[str, Any]
    additionalProperties: bool
    required: List[str] = []
    definitions: Optional[Dict[str, Any]]


class ProcessBaseSchema(OrchestratorBaseModel):
    id: UUID
    workflow_name: str
    product: Optional[UUID]
    customer: Optional[UUID]
    assignee: Assignee
    failed_reason: Optional[str]
    traceback: Optional[str]
    step: Optional[str]
    status: ProcessStatus
    last_step: Optional[str]
    created_by: Optional[str]
    started: datetime
    last_modified: datetime
    subscriptions: List[SubscriptionBaseSchema]
    is_task: bool


class ProcessStepSchema(OrchestratorBaseModel):
    stepid: Optional[UUID]
    name: str
    status: str
    created_by: Optional[str] = None
    executed: Optional[datetime]
    commit_hash: Optional[str] = None
    state: Optional[Dict[str, Any]]


class ProcessSchema(ProcessBaseSchema):
    current_state: Dict[str, Any]
    steps: List[ProcessStepSchema]
    form: Optional[ProcessForm]


class ProcessSubscriptionProcessSchema(OrchestratorBaseModel):
    workflow: str
    pid: UUID
    is_task: bool
    created_by: Optional[str]
    failed_reason: Optional[str]
    started_at: datetime
    last_status: ProcessStatus
    assignee: Assignee
    last_modified_at: datetime
    traceback: Optional[str]
    last_step: Optional[str]

    class Config:
        orm_mode = True


class ProcessSubscriptionBaseSchema(OrchestratorBaseModel):
    workflow_target: Optional[Target]
    subscription_id: UUID
    id: UUID
    pid: UUID
    created_at: datetime

    class Config:
        orm_mode = True


class ProcessSubscriptionSchema(ProcessSubscriptionBaseSchema):
    process: ProcessSubscriptionProcessSchema


class ProcessListItemSchema(OrchestratorBaseModel):
    assignee: Assignee
    created_by: Optional[str]
    failed_reason: Optional[str]
    last_modified_at: datetime
    pid: UUID
    started_at: datetime
    last_status: ProcessStatus
    last_step: Optional[str]
    workflow: str
    workflow_target: Optional[Target]
    is_task: bool
    subscriptions: List[SubscriptionSchema]



import pytest
from pydantic.dataclasses import dataclass

from orchestrator.utils.vlans import VlanRanges


def test_vlanranges_instantiation():
    assert VlanRanges() == VlanRanges([])
    assert VlanRanges(None) == VlanRanges([])
    assert VlanRanges("") == VlanRanges([])
    assert VlanRanges(4) == VlanRanges("4")
    assert VlanRanges("0") == VlanRanges([(0, 0)]) == VlanRanges([[0]])
    assert VlanRanges("2,4,8") == VlanRanges([(2, 2), (4, 4), (8, 8)]) == VlanRanges([[2], [4], [8]])
    assert VlanRanges("80-120") == VlanRanges([(80, 120)]) == VlanRanges([[80, 120]])
    assert VlanRanges("10,12-16") == VlanRanges([(10, 10), (12, 16)]) == VlanRanges([[10], [12, 16]])

    # String interpretation is quite flexible, allowing extra whitespace
    assert VlanRanges("  4   , 6-   10") == VlanRanges("4,6-10")

    # Overlapping ranges will be normalized
    assert VlanRanges("4,6-9,7-10") == VlanRanges("4,6-10")
    assert VlanRanges([[4], [6, 9], [7, 10]]) == VlanRanges([[4], [6, 10]])

    # Non-sensical ranges are not an error perse
    assert VlanRanges("10-1") == VlanRanges("")


def test_vlanranges_str_repr():
    vr = VlanRanges("10-14,4,200-256")

    # `str` version of VlanRanges should be suitable value for constructor, resulting in equivalent object
    assert vr == VlanRanges(str(vr))

    # `repr` version of VlanRanges should be valid Python code resulting in equivalent object
    vr_from_repr = eval(repr(vr), globals(), locals())  # noqa: S307 Use of possibly insecure function
    assert vr_from_repr == vr


def test_vlanranges_in():
    vr = VlanRanges("10-20")
    assert 10 in vr
    assert 20 in vr
    assert 9 not in vr
    assert 21 not in vr
    assert 0 not in vr


def test_vlanranges_intersects():
    vr = VlanRanges("10-20")
    assert not vr & VlanRanges("8-9")
    assert vr & VlanRanges("9-10")
    assert not vr & VlanRanges("21-23")
    assert vr & VlanRanges("20-23")
    assert vr & VlanRanges("10-20")
    assert vr & VlanRanges("11-19")

    assert vr & VlanRanges("0-3,10,20,21-28")
    assert vr & VlanRanges("0-3,20")
    assert not vr & VlanRanges("0-3,9,21,22-30")


def test_vlanranges_sub():
    vr = VlanRanges("10-20")
    assert vr - VlanRanges("8-9") == VlanRanges("10-20")
    assert vr - VlanRanges("9-10") == VlanRanges("11-20")
    assert vr - VlanRanges("21-23") == VlanRanges("10-20")
    assert vr - VlanRanges("20-23") == VlanRanges("10-19")
    assert vr - VlanRanges("10-20") == VlanRanges("")
    assert vr - VlanRanges("11-19") == VlanRanges("10,20")

    assert vr - VlanRanges("0-3,10,20,21-28") == VlanRanges("11-19")
    assert vr - VlanRanges("0-3,20") == VlanRanges("10-19")
    assert vr - VlanRanges("0-3,9,21,22-30") == VlanRanges("10-20")


def test_vlanranges_and():
    vr = VlanRanges("10-20")
    assert vr & VlanRanges("8-9") == VlanRanges("")
    assert vr & VlanRanges("9-10") == VlanRanges("10")
    assert vr & VlanRanges("21-23") == VlanRanges("")
    assert vr & VlanRanges("20-23") == VlanRanges("20")
    assert vr & VlanRanges("10-20") == VlanRanges("10-20")
    assert vr & VlanRanges("11-19") == VlanRanges("11-19")

    assert vr & VlanRanges("0-3,10,20,21-28") == VlanRanges("10,20")
    assert vr & VlanRanges("0-3,20") == VlanRanges("20")
    assert vr & VlanRanges("0-3,9,21,22-30") == VlanRanges("")


def test_vlanranges_or():
    vr = VlanRanges("10-20")
    assert vr | VlanRanges("8-9") == VlanRanges("8-20")
    assert vr | VlanRanges("9-10") == VlanRanges("9-20")
    assert vr | VlanRanges("21-23") == VlanRanges("10-23")
    assert vr | VlanRanges("20-23") == VlanRanges("10-23")
    assert vr | VlanRanges("10-20") == VlanRanges("10-20")
    assert vr | VlanRanges("11-19") == VlanRanges("10-20")

    assert vr | VlanRanges("0-3,10,20,21-28") == VlanRanges("0-3,10-28")
    assert vr | VlanRanges("0-3,20") == VlanRanges("0-3,10-20")
    assert vr | VlanRanges("0-3,9,21,22-30") == VlanRanges("0-3,9-30")


def test_vlanranges_union():
    vr = VlanRanges("10-20")
    # with iterable
    assert vr == VlanRanges().union(VlanRanges("10-15"), [16, 17, 18], (19,), VlanRanges(20))

    # single arg
    assert vr == VlanRanges("10-19").union(VlanRanges(20))


def test_vlanranges_xor():
    vr = VlanRanges("10-20")
    assert vr ^ VlanRanges("8-9") == VlanRanges("8-20")
    assert vr ^ VlanRanges("9-10") == VlanRanges("9,11-20")
    assert vr ^ VlanRanges("21-23") == VlanRanges("10-23")
    assert vr ^ VlanRanges("20-23") == VlanRanges("10-19,21-23")
    assert vr ^ VlanRanges("10-20") == VlanRanges("")
    assert vr ^ VlanRanges("11-19") == VlanRanges("10,20")

    assert vr ^ VlanRanges("0-3,10,20,21-28") == VlanRanges("0-3,11-19,21-28")
    assert vr ^ VlanRanges("0-3,20") == VlanRanges("0-3,10-19")
    assert vr ^ VlanRanges("0-3,9,21,22-30") == VlanRanges("0-3,9-30")


def test_vlanranges_lt():
    vr = VlanRanges("10-20")
    assert not VlanRanges("8-9") < vr
    assert not VlanRanges("9-10") < vr
    assert not VlanRanges("21-23") < vr
    assert not VlanRanges("20-23") < vr
    assert not VlanRanges("10-20") < vr
    assert VlanRanges("11-19") < vr

    assert not VlanRanges("0-3,10,20,21-28") < vr
    assert not VlanRanges("0-3,20") < vr
    assert not VlanRanges("0-3,9,21,22-30") < vr


def test_vlanranges_hash():
    vr = VlanRanges("10-14,4,200-256")
    # Just making sure it doesn't raise an exception. Which, BTW will be raised, should the internal representation
    # be changed to a mutable data structure.
    assert hash(vr)


def test_vlanranges_serialization_deserialization():
    vr = VlanRanges("3,4-10")
    vr2 = VlanRanges(vr.__json__())
    assert vr == vr2


def test_vlanranges_validations():
    # Negative values, however, are an error
    with pytest.raises(ValueError):
        VlanRanges("-10")

    with pytest.raises(ValueError):
        VlanRanges("fubar")

    # Pydantic validation support is present
    @dataclass
    class TestVlanRanges:
        vlanrange: VlanRanges

    assert TestVlanRanges(vlanrange="12").vlanrange == VlanRanges(12)

    with pytest.raises(ValueError):
        TestVlanRanges(vlanrange="bla")

    with pytest.raises(ValueError):
        TestVlanRanges(vlanrange="-30")

    with pytest.raises(ValueError):
        TestVlanRanges(vlanrange="5000")



#!/usr/bin/env python3

# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Optional, Type

import sentry_sdk
import structlog
import typer
from fastapi.applications import FastAPI
from fastapi_etag.dependency import add_exception_handler
from nwastdlib.logging import initialise_logging
from opentelemetry import trace
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor
from opentelemetry.instrumentation.redis import RedisInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.responses import JSONResponse, Response

from orchestrator.api.api_v1.api import api_router
from orchestrator.api.error_handling import ProblemDetailException
from orchestrator.cli.main import app as cli_app
from orchestrator.db import db, init_database
from orchestrator.db.database import DBSessionMiddleware
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY, SubscriptionModel
from orchestrator.exception_handlers import form_error_handler, problem_detail_handler
from orchestrator.forms import FormException
from orchestrator.settings import AppSettings, app_settings, tracer_provider
from orchestrator.version import GIT_COMMIT_HASH

logger = structlog.get_logger(__name__)


class OrchestratorCore(FastAPI):
    def __init__(  # type: ignore
        self,
        title: str = "The Orchestrator",
        description: str = "The orchestrator is a project that enables users to run workflows.",
        openapi_url: str = "/api/openapi.json",
        docs_url: str = "/api/docs",
        redoc_url: str = "/api/redoc",
        version: str = "1.0.0",
        default_response_class: Type[Response] = JSONResponse,
        base_settings: AppSettings = app_settings,
        **kwargs,
    ) -> None:
        super().__init__(
            title=title,
            description=description,
            openapi_url=openapi_url,
            docs_url=docs_url,
            redoc_url=redoc_url,
            version=version,
            default_response_class=default_response_class,
            **kwargs,
        )

        initialise_logging()

        self.include_router(api_router, prefix="/api")

        init_database(base_settings)

        self.add_middleware(SessionMiddleware, secret_key=base_settings.SESSION_SECRET)
        self.add_middleware(DBSessionMiddleware, database=db)
        origins = base_settings.CORS_ORIGINS.split(",")
        self.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_methods=base_settings.CORS_ALLOW_METHODS,
            allow_headers=base_settings.CORS_ALLOW_HEADERS,
            expose_headers=base_settings.CORS_EXPOSE_HEADERS,
        )

        self.add_exception_handler(FormException, form_error_handler)
        self.add_exception_handler(ProblemDetailException, problem_detail_handler)
        add_exception_handler(self)

        @self.router.get("/", response_model=str, response_class=JSONResponse, include_in_schema=False)
        def _index() -> str:
            return "Orchestrator orchestrator"

    def instrument_app(self) -> None:
        logger.info("Activating Opentelemetry tracing to app", app=self.title)
        trace.set_tracer_provider(tracer_provider)
        FastAPIInstrumentor.instrument_app(self)
        RequestsInstrumentor().instrument()
        RedisInstrumentor().instrument()
        Psycopg2Instrumentor().instrument()
        SQLAlchemyInstrumentor().instrument(engine=db.engine, tracer_provider=tracer_provider)

    def add_sentry(
        self,
        sentry_dsn: str,
        trace_sample_rate: float,
        server_name: str,
        environment: str,
        release: Optional[str] = GIT_COMMIT_HASH,
    ) -> None:
        logger.info("Adding Sentry middelware to app", app=self.title)
        sentry_sdk.init(
            dsn=sentry_dsn,
            traces_sample_rate=trace_sample_rate,
            server_name=server_name,
            environment=environment,
            release=f"orchestrator@{release}",
            integrations=[SqlalchemyIntegration(), RedisIntegration()],
        )
        self.add_middleware(SentryAsgiMiddleware)

    @staticmethod
    def register_subscription_models(product_to_subscription_model_mapping: Dict[str, Type[SubscriptionModel]]) -> None:
        """
        Register your subscription models.

        This method is needed to register your subscription models inside the orchestrator core.

        Args:
            product_to_subscription_model_mapping: The dictionary should contain a mapping of products to SubscriptionModels.
                The selection will be done depending on the name of the product.

        Returns:
            None:

        Examples:
            product_to_subscription_model_mapping = {
                "Generic Product One": GenericProductModel,
                "Generic Product Two": GenericProductModel,
            }

        """
        SUBSCRIPTION_MODEL_REGISTRY.update(product_to_subscription_model_mapping)


main_typer_app = typer.Typer()
main_typer_app.add_typer(cli_app, name="orchestrator", help="The are the orchestrator cli commands")

if __name__ == "__main__":
    main_typer_app()



import pytest
import schedule

from orchestrator.cli.scheduler import run
from orchestrator.schedules import ALL_SCHEDULERS
from orchestrator.schedules.scheduling import scheduler


def test_scheduling_with_period(capsys, monkeypatch):

    ref = {"called": False}

    @scheduler(name="test", time_unit="second", period=1)
    def test_scheduler():
        ref["called"] = True
        print("I've run")  # noqa: T001
        return schedule.CancelJob

    ALL_SCHEDULERS.clear()
    ALL_SCHEDULERS.append(test_scheduler)

    # Avoid having to mock next_run() and idle_seconds() deep in the scheduler as we are only interested in the job:
    with pytest.raises(TypeError):
        run()
        captured = capsys.readouterr()
        assert captured.out == "I've run\n"
        assert ref["called"]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

SPEED_TO_INTERFACE_TYPE: Dict[int, List[str]] = {
    1000: ["1000BASE-T", "1000BASE-EX", "1000BASE-LX", "1000BASE-SX", "1000BASE-ZX", "1000BASE-CWDM"],
    10000: ["10GBASE-ER", "10GBASE-LR", "10GBASE-SR", "10GBASE-ZR", "10GBASE-CWDM"],
    40000: ["40GBASE-LR4", "40GBASE-SR4", "40GBASE-CWDM"],
    100000: ["100GBASE-LR4", "100GBASE-SR4", "100GBASE-LR10", "100GBASE-SR10", "100GBASE-CWDM4"],
}

INTERFACE_TYPE_TO_SPEED: Dict[str, int] = {
    ieee_type: speed for speed, ieee_types in SPEED_TO_INTERFACE_TYPE.items() for ieee_type in ieee_types
}


SPEED_TO_IMS_SPEED: Dict[int, str] = {
    1000: "1G",
    10000: "10G",
    40000: "40G",
    100000: "100G",
}



from http import HTTPStatus
from uuid import uuid4

import pytest

from orchestrator.config import IMS_CIRCUIT_ID
from orchestrator.db import ProductBlockTable, ProductTable, ResourceTypeTable, db
from orchestrator.utils.json import json_dumps

PRODUCT_BLOCK_ID = "f51f9542-e83f-42e5-a590-0284dd5493e4"


@pytest.fixture
def seed():
    ProductBlockTable.query.delete()
    resources = [ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="IMS Circuit Id")]
    product_blocks = [
        ProductBlockTable(
            product_block_id=PRODUCT_BLOCK_ID,
            name="Ethernet",
            description="desc",
            status="active",
            resource_types=resources,
            tag="tag",
        )
    ]
    product = ProductTable(
        name="ProductTable",
        description="description",
        product_type="Port",
        tag="Port",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=[],
    )

    db.session.add(product)
    db.session.commit()


def test_fetch_all_product_blocks(seed, test_client):
    response = test_client.get("/api/product_blocks")

    assert HTTPStatus.OK == response.status_code
    product_blocks = response.json()

    assert 1 == len(product_blocks)


def test_get_by_id(seed, test_client):
    product_block = test_client.get(f"/api/product_blocks/{PRODUCT_BLOCK_ID}").json()
    assert IMS_CIRCUIT_ID == product_block["resource_types"][0]["resource_type"]


def test_save_product_block(seed, test_client):
    body = {"name": "Circuit", "description": "desc", "status": "active"}
    response = test_client.post(
        "/api/product_blocks/", data=json_dumps(body), headers={"Content_Type": "application/json"}
    )
    assert HTTPStatus.NO_CONTENT == response.status_code
    assert 2 == len(test_client.get("/api/product_blocks").json())


def test_save_product_block_with_existing_resource_type(seed, test_client):
    resource_type = test_client.get("/api/resource_types").json()[0]
    body = {"name": "Circuit", "description": "desc", "status": "active", "resource_types": [resource_type]}
    response = test_client.post(
        "/api/product_blocks/", data=json_dumps(body), headers={"Content_Type": "application/json"}
    )

    assert HTTPStatus.NO_CONTENT == response.status_code
    assert 2 == len(test_client.get("/api/product_blocks").json())


def test_save_invalid_product_block(seed, test_client):
    response = test_client.post(
        "/api/product_blocks/", data=json_dumps({}), headers={"Content_Type": "application/json"}
    )

    assert HTTPStatus.UNPROCESSABLE_ENTITY == response.status_code
    assert {
        "detail": [
            {"loc": ["body", "name"], "msg": "field required", "type": "value_error.missing"},
            {"loc": ["body", "description"], "msg": "field required", "type": "value_error.missing"},
        ]
    } == response.json()


def test_update_product_block(seed, test_client):
    body = {"product_block_id": PRODUCT_BLOCK_ID, "name": "Circuit", "description": "desc", "status": "end of life"}
    response = test_client.put(
        "/api/product_blocks/", data=json_dumps(body), headers={"Content_Type": "application/json"}
    )
    assert HTTPStatus.NO_CONTENT == response.status_code
    product_block = test_client.get(f"/api/product_blocks/{PRODUCT_BLOCK_ID}").json()
    assert "end of life" == product_block["status"]


def test_delete_product_block(seed, test_client):
    pb_id = str(uuid4())
    body = {"product_block_id": pb_id, "name": "Circuit", "description": "desc", "status": "active"}
    test_client.post("/api/product_blocks/", data=json_dumps(body), headers={"Content_Type": "application/json"})

    response = test_client.delete(f"/api/product_blocks/{pb_id}")
    assert HTTPStatus.NO_CONTENT == response.status_code


def test_delete_product_block_with_used_products(seed, test_client):
    response = test_client.delete(f"/api/product_blocks/{PRODUCT_BLOCK_ID}")
    assert HTTPStatus.BAD_REQUEST == response.status_code

    assert f"ProductBlock {PRODUCT_BLOCK_ID} is used in Products: ProductTable" == response.json()["detail"]



import pytest

from orchestrator.db import SubscriptionTable, WorkflowTable, db
from orchestrator.forms import FormValidationError
from orchestrator.targets import Target
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.utils import validate_workflow
from test.unit_tests.workflows import (
    WorkflowInstanceForTests,
    assert_complete,
    assert_failed,
    extract_error,
    run_workflow,
)


@validate_workflow("Test Validation")
def validation_workflow() -> StepList:
    return StepList([])


WorkflowInstanceForTests(validation_workflow, "validation_workflow")


def test_happy_flow(generic_subscription_1):
    product = SubscriptionTable.query.get(generic_subscription_1).product
    product.workflows.append(WorkflowTable(name="validation_workflow", target=Target.SYSTEM))
    db.session.add(product)
    db.session.commit()

    result, process, step_log = run_workflow("validation_workflow", {"subscription_id": generic_subscription_1})
    assert_complete(result)


def test_no_subscription():
    with pytest.raises(FormValidationError) as error_info:
        run_workflow("validation_workflow", {"subscription_id": None})
    assert "none is not an allowed value" in str(error_info.value)


def test_failed_validation(generic_subscription_1: str) -> None:
    @step("Fail")
    def fail():
        raise ValueError("Failed")

    @validate_workflow("Failing Validation")
    def failing_validation_workflow() -> StepList:
        return begin >> fail

    WorkflowInstanceForTests(failing_validation_workflow, "failing_validation_workflow")

    product = SubscriptionTable.query.get(generic_subscription_1).product
    product.workflows.append(WorkflowTable(name="failing_validation_workflow", target=Target.SYSTEM))
    db.session.add(product)
    db.session.commit()

    result, process, step_log = run_workflow("failing_validation_workflow", {"subscription_id": generic_subscription_1})
    assert_failed(result)
    assert "Failed" in extract_error(result)
